#!/usr/bin/env python
# -*- coding: UTF-8 -*-

""" Script pour l'installation d'une distrib dans un domU de xen """

import commands
import sys
import os
from creole import parsedico
import subprocess
import time
import glob

DICO = parsedico.parse_dico()
ADRESSE_IP_DMZ = DICO['adresse_ip_dmz']
SERVEUR_MAJ = DICO['serveur_maj']

def set_auto_start(distrib):
    """ Remplis le fichier de config du domU avec le nom du noyau
    Relecture du fichier /etc/xen/xen-* pour changer le nom du noyau et du ramdisk
    A revoir plus tard
    """

    ret = run('apt-cache show linux-image-eolexen | grep "^Depends" | head -n1 | awk "{ print \$2 }"|sed "s/,//g" | sed -e "s/linux-image-//" | tr -d "[[:space:]]"')
    if ret :
        run('sed -e "s:NOYAU:/boot/vmlinuz-%s:" -e "s:RAMDISK:/boot/initrd.img-%s:" -e "s:DISTRIB:%s:g" /usr/share/eole/domu/conf/xen-%s > /etc/xen/xen-%s' % (ret, ret, distrib, distrib, distrib))
#        run('sed -e "s:NOYAU:/boot/vmlinuz-%s:" -e "s:RAMDISK:/boot/initrd.img-%s:" /etc/xen/xen-%s' % (ret, ret, distrib))
        run('ln -sf /etc/xen/xen-%s /etc/xen/auto' % distrib)


def make_partitions(config, distrib):
    """
    Teste si une partition est montée si c'est le cas,
    récupère le device sur lequel est montée la partition
    démonte la partition nettoie le fichier /etc/fstab
    crée un volume physique
    crée le volume groupe
    teste la présence du VG xen-eole
    teste si le VG xen-eole est assez grand
    crée les Volumes Logiques
    """
    partitions = config["PARTITIONS"]
    xen_device = run("mount | grep /xen |awk {'print $1'}")
    ret = commands.getstatusoutput("umount /xen >/dev/null")
    if ret[0] == 0 :
        #nettoyage du fichier fstab
        old_xen_fstab = run("grep -ni /xen /etc/fstab |cut -d : -f 1")
        line = int(old_xen_fstab) - 1
        run('sed -e %d,+1d /etc/fstab >/tmp/fstab' % line)
        run('cp /tmp/fstab /etc/fstab')
        run('rm /tmp/fstab')
        # création du volume physique
        run("pvcreate %s" % xen_device)
        # création du volume logique
        run("vgcreate xen-eole %s" % xen_device)
    # teste la présence du vg xen-eole
    run("vgdisplay xen-eole 2>/dev/null | grep 'VG Name' >/dev/null")
    # teste si le vg xen-eole est assez grand
    vg_size = run("vgdisplay xen-eole 2>/dev/null| grep -iE 'Free *PE' | awk {'print $5'}")
    if int(vg_size) < 2500 :
        print "le volume group n'a pas assez d'espace disponible( moins de 10G )"
        print "Vous devez modifier la taille de vos partitions"
        sys.exit('Place insuffisante')
#        remove_logical_volume(distrib)
    else:
        print "le volume group xen-eole est suffissement grand %s PE" % vg_size
    if not os.path.isdir('/xen') :
        run("mkdir /xen")
    for logic_vol in partitions :
        # le point de montage est la valeur
        lv_mount_point = logic_vol[0]
        name =  logic_vol[0].replace('/','')
        if name == "" :
            name = "root"
        lv_name = "%s-%s" % (distrib, name)
        lv_size = logic_vol[1]
        if lv_size == "*":
            lv_size = run("vgdisplay xen-eole 2>/dev/null | grep -iE 'Free *PE' | awk '{ print $5*4 }'")
        ret = commands.getstatusoutput("lvdisplay /dev/xen-eole/%s 2>/dev/null " % lv_name)
        if ret[0] != 0:
            print "création du volume Logique %s de taille %s pour le point de montage %s" % (lv_name, lv_size, lv_mount_point)
            run('lvcreate -n %s -L %s xen-eole 2>/dev/null ' % (lv_name, lv_size))
        else :
            print "le volume logique %s existe" % lv_name
        if name != "swap" :
            print "formatage du volume Logique %s en EXT3" % lv_name
            run("mkfs.ext3 -q /dev/xen-eole/%s 2>/dev/null " % lv_name)
            print "création du point de montage %s pour le Volume Logique %s" % (lv_mount_point, lv_name)
            run("mkdir -p /xen/%s" % lv_mount_point)
            print "montage du volume logique %s sur le point de montage %s" % (lv_name, lv_mount_point)
            run("mount /dev/xen-eole/%s /xen/%s 2>/dev/null " % (lv_name, lv_mount_point))
        else :
            print "activation de l'espace de SWAP"
            run("mkswap /dev/xen-eole/%s 2>/dev/null " % lv_name)

def remove_logical_volume(distrib):
    if commands.getstatusoutput('/usr/sbin/xm list | grep xen-%s' % distrib)[0] == 0 :
        run('/usr/sbin/xm destroy xen-%s' % distrib)
        print "attente de l'arrêt du domaine xen-%s ..." % distrib
        time.sleep(10)
    lv_to_remove = run("lvdisplay 2>/dev/null|grep 'LV Name' | grep %s- | awk '{print $3}'" % distrib).split('\n')
    if commands.getstatusoutput("mount |grep /dev/mapper/xen--eole-%s--root" % distrib)[0] == 0:
        run_subprocess_without_exit(['chroot', '/xen', 'umount', '-a'])
        umount_partitions()

    if commands.getstatusoutput("mount |grep /dev/mapper/xen--eole-%s--root" % distrib)[0] == 0:
        sys.exit("veuillez redémarrer le serveur et relancer le script d'installation")
    else:
        for lvs in lv_to_remove :
            run_subprocess(['lvremove', '-f',"%s" % lvs])

def check_cdrom_device():
    """testons si le device du cdrom est accessible"""
    real_cd_device = run("cat /proc/sys/dev/cdrom/info |grep \"drive name\" | awk '{ print $NF }'")
    cdrom_device = "/dev/"+real_cd_device
    cdrom_mount_point = run("grep cdrom /etc/fstab | awk {'print $2'}")
    if not os.path.isdir(cdrom_mount_point):
        run("mkdir %s" % cdrom_mount_point)
    #run('/sbin/modprobe sr_mod')
    #cdrom_device = run("grep cdrom /etc/fstab | awk {'print $1'}")
    #if not os.path.exists(cdrom_device):
    #    run("/bin/mknod -m 660 %s b 11 0" % cdrom_device)
    #    run("chown root:cdrom %s" % cdrom_device)
    return(cdrom_device, cdrom_mount_point)

def check_eole_cd():
    """fonction de détection du cdrom EOLENG
    création du point de montage dans le chroot et montage du cdrom"""

    cd_device, cd_mount_point = check_cdrom_device()
    print "Veuillez insérer le cdrom Eole NG dans le lecteur et validez"
    if raw_input() in ["ok", "ok", ""] :
        ret = commands.getstatusoutput("mount -o ro %s %s" % (cd_device, cd_mount_point))
        if ret[0] != 0 :
            print "Erreur lors du montage du CD"
            return(1)
        else :
            if os.path.isfile("%s/preseed/amonng.seed" % cd_mount_point):
                print "Le cdrom Eole a été trouvé."
                return(0)
            else :
                return(1)
    else :
        return(1)

def mount_chroot_cd():
    """montage du cdrom dans le chroot"""

    # création du point de montage pour le cdrom dans le chroot du domU
    run("mkdir /xen/cdrom")
    cdrom_mount_point = run("grep cdrom /etc/fstab | awk {'print $2'}")
    # montage du cdrom dans l'environnement chroot
    run("mount --bind %s /xen/cdrom" % (cdrom_mount_point))

def chroot_xen(source_install, distrib):
    """chroot dans l'environnement du domU"""

    print "rentrée dans le chroot"
    run("cp /usr/share/eole/domu/scripts/prepare_domu.py /xen/root/")
    run("chmod a+x /xen/root/prepare_domu.py")
    cmd = ['chroot', '/xen', '/usr/bin/python', '/root/prepare_domu.py',source_install, distrib]
    run_subprocess(cmd)


def umount_chroot_cd():
    """démontage du cdrom eole"""

    print "démontage du cdrom dans le chroot /xen"
    run("umount /xen/cdrom")

def umount_cd():
    """démontage du cdrom"""

    print "démontage du cdrom"
    cdrom_mount_point = run("grep cdrom /etc/fstab | awk {'print $2'}")
    run("umount %s" % cdrom_mount_point)

def debootstrap_install(source):
    """
    fonction d'installation minimum
    soit depuis le réseau soit depuis le cdrom
    reseau : http://eoleng.ac-dijon.fr/ubuntu/
    cdrom : file:///media/cdrom0/
    """

    if source == "reseau" :
    #    source = "http://eoleng.ac-dijon.fr/ubuntu/"
        source = "http://%s/ubuntu/" % SERVEUR_MAJ
    else :
        source = "file:///media/cdrom0/"

    print "récupération via debootstrap depuis %s" % source
    cmd = ['debootstrap', '--components=main,local,restricted,universe',
            '--include=libc6-xen,ubuntu-keyring,ubuntu-minimal,openssh-server', 'gutsy', '/xen', '%s' % source]
    run_subprocess(cmd)


def generate_ssh_rsa_key():
    """génération de la clef rsa dans le domU pour les connexions SSH"""

    print "génération de la clef rsa dans le domU"
    if not os.path.isfile('/root/.ssh/id_rsa'):
        run_subprocess(['ssh-keygen', '-t', 'rsa', '-P', '', '-f', '/root/.ssh/id_rsa'])
    run('mkdir -p /xen/root/.ssh/ 2>/dev/null')
    run('cat /root/.ssh/id_rsa.pub > /xen/root/.ssh/authorized_keys')
    run('chmod 700 /xen/root/.ssh/authorized_keys')


def set_domu_timezone(distrib):
    """réglage de la timezone du domu"""

    print "reglage de la timezone du domU"
    run("cp /etc/timezone /xen/etc/timezone")

def set_domu_network_config(distrib):
    """réglage de la configuration réseau"""

    print "configuration du réseau du domU"
    address = DICO['ip_xen_%s' % distrib]
    netmask = DICO['netmask_xen_%s' % distrib ]
    gw = DICO['gw_xen_%s' % distrib ]
    nom_domaine_local = DICO['nom_domaine_local']
    network_interfaces_file = ['auto lo', 'iface lo inet loopback', ' ',
            'auto eth0', 'iface eth0 inet static', 'address %s' % address,
            'netmask %s' % netmask,  'gateway %s' % gw,
            'post-up /sbin/ifconfig eth0 mtu 1496',
            'post-up /usr/sbin/ethtool -K eth0 tx off']

    write_file('/xen/etc/network/interfaces', network_interfaces_file)

    resolv_conf= ['nameserver %s' % gw,'search %%nom_domaine_local']
    write_file('/xen/etc/resolv.conf', resolv_conf)

    # le petit truc pour que le reseau demarre au boot
#    run("cp /etc/rc.local /etc/rc.local.orig")
#    run("sed -i 's/\/usr\/sbin\/ethtool \-K eth0 tx off//g' /etc/rc.local")
#    run("sed -i 's/exit 0//g' /etc/rc.local")
#    run("echo '/usr/sbin/ethtool -K eth0 tx off' >> /etc/rc.local")
#    run("echo 'exit 0' >> /etc/rc.local")
#    run("cat /etc/rc.local.orig | sed 's:^exit 0:\/usr\/sbin\/ethtool \-K eth0 tx off\\nexit 0:'>/etc/rc.local")
#    run("cp /etc/rc.local /xen/etc/rc.local")

def set_domu_fstab(config, distrib):
    """mise en place du fichier /etc/fstab"""

    print "configuration du fichier /etc/fstab dans le domU"
    fstab_file = ['# /etc/fstab: static file system information.',
            '#', '# <file system> <mount point>   <type>  <options>       <dump>  <pass>'] + config["FSTAB"]
    write_file('/xen/etc/fstab', fstab_file)

def set_domu_sources_list(source_installation):
    print "Mise en place du sources.list pour %s" % source_installation
    if source_installation == "cdrom":
        sources_list = "deb file:///cdrom gutsy main restricted universe local"
        run('echo "%s" > /xen/etc/apt/sources.list' % sources_list)
    else:
        sources_list = ["deb http://%s/ubuntu/ gutsy main restricted universe" % SERVEUR_MAJ,
                "deb http://%s/ubuntu/ gutsy-security main restricted universe" % SERVEUR_MAJ,
                "deb http://%s/ubuntu/ gutsy-updates main restricted universe" % SERVEUR_MAJ,
                "deb http://%s/eoleng/ eole-2.1/" % SERVEUR_MAJ,
                "deb http://%s/eoleng/ eole-2.1-updates/" % SERVEUR_MAJ,
                "deb http://%s/eoleng/ eole-2.1-security/" % SERVEUR_MAJ
                ]

        for source in sources_list:
            if sources_list.index(source) == 0 :
                run("echo %s > /xen/etc/apt/sources.list" % source)
            else:
                run("echo %s >> /xen/etc/apt/sources.list" % source)


def umount_partitions():
    """Démontage des partitions"""

    print "Démontage des partitions"
    partitions = run("df -P | grep xen | awk '{ print $6 }' | sort -r 2>/dev/null")
    liste_partitions = partitions.split("\n")
    for part in liste_partitions :
        run_subprocess_without_exit(['umount', part])

def write_file(output_file, file_content):
    """ Ecrire file_content dans output_file """

    try:
        filehandle = file(output_file, 'w')
        filehandle.write('\n'.join(file_content))
        filehandle.close()
    except Exception:
        print "Impossible d'ecrire dans le fichier %s" % output_file

def modify_iftab():
    """eth0 en derniere position dans /etc/iftab"""
    if os.path.isfile('/etc/iftab'):
        iftab_orig = run('grep eth /etc/iftab').split('\n')
        if int(iftab_orig[0][3]) < int(iftab_orig[1][3]):
            iftab_orig.append(iftab_orig[0])
            del(iftab_orig[0])
            write_file('/etc/iftab', iftab_orig)
        else:
            pass


def run(cmd):
    """raccourci d'execution d'une commande shell"""

    ret, res = commands.getstatusoutput(cmd)
    if ret != 0 :
        print "\033[31mLa commande %s a echouee\033[0m" % cmd
        sys.exit(res)
    return res

def run_subprocess(cmd):
    """Raccourci d'execution pour l'execution d'une commande shell avec capture de STDOUT
    Attention, cmd est une liste :
    cmd='ls -al' s'ecrit cmd=['ls','-la']
    """
    returncode = run_subprocess_without_exit(cmd)
    if returncode:
        print "\033[31mLa commande %s a echoue\033[0m" % cmd
        sys.exit(returncode)

def run_subprocess_without_exit(cmd):
    process = subprocess.Popen(cmd, stdout=sys.stdout, stderr=sys.stderr, stdin=sys.stdin)
    process.communicate()
    return (process.returncode)

def read_config(conf_file):
    """Lire un fichier de configuration"""
    store={}
    try:
        input_file = open(conf_file,'r')
    except Exception:
        print "Impossible de lire le fichier %s" % conf_file

    raw_config = input_file.readlines()
    input_file.close()
    for lines in raw_config:
        line = lines.lstrip()
        if line:
            if line[0] == "[" :
                #Une section
                sec = line[1:len(line)-2].strip()
                store[sec]=[]
            elif line[0].strip() == '\n' or line[0].strip() == '#':
                #Ces lignes sont des commentaires
                continue
            else:
                if sec == 'FSTAB':
                    tab=line.strip()
                else:
                #Les clefs et les valeurs de la section en cours
                    tab=line.strip().replace(' ','',line.count(' ')).split('=')
                store[sec].append(tab)
    if 'PARTITIONS' not in store.keys():
        sys.exit("Le fichier de configuration n'est pas valide")
    else:
        return(store)

def choose_remove_lvm (distrib):
    """
        Demande si l'on veut supprimer l'ensemble des partitions lvm d'une distribution
    """
    ret = commands.getstatusoutput("lvdisplay /dev/xen-eole/%s-root 2>/dev/null " % distrib)
    if ret[0] == 0:
        print ""
        print "Le serveur virtuel \"%s\" existe" % distrib
        print "Voulez-vous le supprimer (O/n) ?"
        print "attention, toutes les données seront perdues"
        if raw_input() in ["o", "O", ""]:
            remove_logical_volume (distrib)

def stopstart_xendomains(action):
    if action == 'stop':
        print "Arret de xendomains"
        cmd = (['/etc/init.d/dhcp3-relay', 'stop'])
        cmd = (['/etc/init.d/bind9', 'stop'])
        cmd = (['/etc/init.d/xendomains', 'stop'])
        FNULL = open('/dev/null', 'w')
        process = subprocess.Popen(cmd, stdout=FNULL, stderr=FNULL, stdin=FNULL)
        process.communicate()
    if action == 'start':
        run("chmod +x /etc/xen/scripts/*")
        run_subprocess_without_exit(['/etc/xen/scripts/network-virtual', 'bridgeip="%s/24"' % DICO['adresse_ip_dmz'],  'brnet="%s/24"' % DICO['adresse_network_dmz'], 'netmask="%s"' % DICO['adresse_netmask_dmz'], 'bridge="%s"' % DICO['nom_dmz'], 'start'])
        run_subprocess_without_exit(['/etc/init.d/xendomains', 'start'])
        run_subprocess_without_exit(['/etc/init.d/dhcp3-relay', 'start'])
        run_subprocess_without_exit(['/etc/init.d/bind9', 'start'])
        
def create_xen (distrib, install_from_media):
    print ''
    print 'Création de %s' % (distrib) 
    if not os.path.isfile('/usr/share/eole/domu/conf/xen-%s' % distrib):
        sys.exit('Pas de fichier de démarrage pour %s' % distrib)
    if not os.path.isfile('/usr/share/eole/domu/conf/%s.conf' % distrib):
        sys.exit('Aucun fichier de configuration pour %s' % distrib)
    if DICO['activate_xen_%s' % distrib] != 'oui':
        if os.path.isfile('/etc/xen/auto/xen-%s' % distrib):
            if commands.getstatusoutput('/usr/sbin/xm list | grep xen-%s' % distrib)[0] == 0 :
                print 'Arrêt de %s' % distrib
                run ('xm shutdown xen-%s' % distrib)
            print 'Suppression du démarrage automatique de %s' % distrib
            run ('rm -f /etc/xen/auto/xen-%s' % (distrib))
        else:
            print 'Le serveur %s virtuel est désactivé. Pour l\'activer, utilisez gen_config.' % distrib
    else:
        config = read_config('/usr/share/eole/domu/conf/%s.conf' % distrib)
        set_auto_start(distrib)
        ret = commands.getstatusoutput("lvdisplay /dev/xen-eole/%s-root 2>/dev/null " % distrib)
        if ret[0] != 0:
            make_partitions(config, distrib)

            if install_from_media == "cdrom":
                mount_chroot_cd()
    
            debootstrap_install(install_from_media)

            set_domu_sources_list(install_from_media)
            chroot_xen(install_from_media, distrib)
            generate_ssh_rsa_key()
            set_domu_timezone(distrib)
            set_domu_network_config(distrib)
            set_domu_fstab(config, distrib)

            if install_from_media == 'cdrom':
                umount_chroot_cd()
            umount_partitions()
        else:
            print 'Le serveur %s est déjà installé, ne fait rien' % distrib

def main():
    """Fonction main"""

    stopstart_xendomains('stop')
    print "quelle méthode d'installation voulez-vous ?"
    rep = ''
    while rep not in ['reseau', 'cdrom']:
        rep = raw_input("Taper [reseau] pour installer depuis le réseau, taper [cdrom] pour installer depuis le cdrom\n")
        if rep == 'cdrom':
            if check_eole_cd() != 0:
                rep = 'none'

    install_from_media = rep
    if len(sys.argv) == 2 :
        distrib = sys.argv[1]
        choose_remove_lvm (distrib)
        create_xen (distrib, install_from_media)
    else:
#        conffiles = glob.glob('/usr/share/eole/domu/conf/xen-*')
	conffiles = ['eclair', 'eolebase', 'scribe']
        for distrib in conffiles:
#            distrib = os.path.split(conffile)[1].replace('xen-', '', 1)
            choose_remove_lvm (distrib)
        for distrib in conffiles:
#            distrib = os.path.split(conffile)[1].replace('xen-', '', 1)
            create_xen (distrib, install_from_media)
    if install_from_media == 'cdrom':
        umount_cd()
    stopstart_xendomains('start')

if __name__ == '__main__':
    main()

